program Back_Propagation;

(*********************************************************
  Backpropagation Network
  Written by Maureen Caudill 
  in Lightspeed C on a Macintosh
  
   Adapted for Quick C by Harris Epstein  Compuserv ID 71121,245

   Adapted for Turbo Pascal by Daniel Orme  CompuServ ID 70132,2730

  This network will accept 5x7 input patterns
  and produce 8 bit output patterns.
  See the June, 1988 AI Expert, Neural Network Primer, Part 3 
  for a discussion of Backpropagation networks.

  *********************************************************)
  
(*********************************************************
  Structure of this program:

  Global data declarations

  Functions:
  	do_forward_pass     propagate input activity forward thru network
  	  do_out_forward      do output layer, forward pass
  	  do_mid_forward      do middle layer, forward pass
  	  display_output      display output of network
  	do_back_pass        propagate error activity backward thru network
  	  do_out_error        compute output layer errors
  	  do_mid_error       compute middle layer errors
  	  adjust_out_wts     adjust output layer weights
  	  adjust_mid_wts    adjust middle layer weights
  	check_out_error    check to see if network knows all patterns yet
  	initialize_net      do network initialization
  	  randomize_wts     randomize wts on middle & output layers
  	  read_data_file      read input/desired out patterns from data file
  	display_mid_wts   output the weights on the middle layer neurodes
  	display_out_wts    output the weights on the output layer neurodes
  	main                     main control program
**********************************************************)

(**********************************************************
        GLOBAL DATA & CONSTANT DECLARATIONS
**********************************************************)
uses
  crt, dos;

const
  T = 1;
  F = 0;
  OutPath = 'BACKPROP.OUT';
  {InPath = 'ALPHABET';}
  {OutWeight = '.OUT';}

  Err = -1;
  MaxPats = 16;                 (* max number of patterns in data file *)
  InXSize = 5;                  (* number of neurodes/row of input layer *)
  InYSize = 7;                  (* number of neurodes/col of input layer *)
  InSize  = InXSize * InYSize;  (* equals IN_X_SIZE*IN_Y_SIZE *)
  MidSize = 4;                  (* number of neurodes in middle layer *)
  OutSize = 8;                  (* number of neurodes in output layer *)
  Margin  = 0.1;                (* how near to 1,0 do we have to come to stop? *)
  Beta    = 0.3;                (* beta learning constant *)
  Alpha   = 0;                  (* momentum term constant *)

type
  OutAryType = array [ 1..OutSize ] of real{**};
  MidWeightsArray = array [ 1..MidSize, 1..InSize ] of real{**};
  OutWeightsArray = array [ 1..OutSize, 1..MidSize ] of real{**};
  MidOutArray     = array [ 1..MidSize ] of real{**};
  OutOutArray     = OutAryType;
  InPatternsArray = array [ 1..MaxPats, 1..InSize ] of real{**};
  OutPatternsArray= array [ 1..MaxPats, 1..OutSize ] of real{**};
  TotalErrorArray = array [ 1..MaxPats ] of real{**};
  Mode = ( Learn, Match );

var
  Mid_Wts   : MidWeightsArray;      (* middle layer weights *)
  Out_Wts   : OutWeightsArray;      (* output layer weights *)
  Mid_Out   : MidOutArray;          (* middle layer output *)
  Out_Out   : OutOutArray;          (* output layer output *)
  Mid_Error : MidOutArray;          (* middle layer errors *)
  Out_Error : OutOutArray;          (* output layer errors *)

  Mid_Wt_Change : MidWeightsArray;  (* storage for last wt change *)
  Out_Wt_Change : OutWeightsArray;  (* storage for last wt change *)
  In_Pats       : InPatternsArray;  (* input patterns *)
  Out_Pats      : OutPatternsArray; (* desired output patterns *)
  Tot_Out_Error : TotalErrorArray;  (* measure of whether net is done *)

  IterationCount,                 (* number of passes thru network so far *)
  NumPats : integer;              (* number of patterns in data file *)
  Learned : boolean;              (* flag--if TRUE, network has learned all patterns *)

  InFile, OutFile, WeightOut : text;
  CurrentMode : Mode;
  OutWeight : string;

(*********************************************************
   do_mid_forward(patt)
   	 process the middle layer's forward pass
   	 The activation of middle layer's neurode is the weighted
   	 sum of the inputs from the input pattern, with sigmoid
   	 function applied to the inputs.
**********************************************************)
procedure Do_Mid_Forward( var Patt : byte );

  var
    Sum, Signal, Activity, InSig : real{**};
    Neurode, i : integer;

  begin { Do_Mid_Forward }
    for Neurode := 1 to MidSize do
      begin
	Sum := 0.0;
	for i := 1 to InSize do  (* compute weighted sum of input signals *)
	  Sum := Sum + ( Mid_Wts[ Neurode, i ] * In_Pats[ Patt, i ] );
	(* apply function f(x) = 1/(1+exp(-5x)) to weighted sum       *)
	(* the factor of 5 is arbitrarily chosen to give a good range *)
	(* from almost 0 to nearly 1                                  *)
	Sum := 1 / (1 + exp( -5 * Sum ) );
	Mid_Out[ Neurode ] := Sum;
      end;
  end; { Do_Mid_Forward }

(**********************************************************
	do_out_forward()
	  process the forward pass through the output layer
	  The activation of the output layer is the weighted sum of
	  the inputs (outputs from middle layer), modified by the
	  sigmoid function.
**********************************************************)
procedure Do_Out_Forward;

  var
    Sum, Signal, Activity, InSig : real{**};
    Neurode, i : integer;

  begin { Do_Out_Forward }
    for Neurode := 1 to OutSize do
      begin
        Sum := 0.0;
	for i := 1 to MidSize do  (* compute weighted sum of input signals from middle layer *)
	  Sum := Sum + ( Out_Wts[ Neurode, i ] * Mid_Out[ i ] );
	(* apply f(x) = 1/(1+exp(-5x)) to weighted input              *)
	(* the factor of 5 is arbitrarily chosen to give a good range *)
	(* from almost 0 to nearly 1                                  *)
	Sum := 1/( 1 + exp( -5 * Sum ));
        Out_Out[ Neurode ] := Sum;
        {Out_Out[ Neurode ] := round( Sum );}
      end;
  end; { Do_Out_Forward }

function Out_Val( var OutAry : OutAryType ): Char;

  var
    ch : byte;
    i : byte;
    x : real;

  begin { Out_Val }
    ch := 0;
    for i := 1 to OutSize do
      if round( OutAry[ i ] ) = 1 then
        begin
          x := exp( (OutSize-i) * ln( 2 ) );
          inc( ch, trunc( x ) );
        end;
    Out_Val := char( ch );
  end; { Out_Val }

(*******************************************************
	display_output(patt)
		Display the actual output vs. the desired output of the
		network.
		Once the training is complete, and the "learned" flag set
		to TRUE, then display_output sends its output to both
		the screen and to a text output file.
********************************************************)
procedure Display_Output( var Patt : byte; Matching : boolean );

  var
    i : integer;

  procedure Do_Report( var Patt : byte );

    begin { Do_Report }
        writeln( {Patt:6}' ':6, ' ':7,
                 Out_Val( OutAryType( Out_Pats[ Patt ] ) ),' ':8,
                 Out_Val( Out_Out ) );
      writeln( OutFile );
    end; { Do_Report }

  begin { Display_Output }
    if Matching then
      Do_Report( Patt );
    writeln( OutFile );
    writeln( OutFile, 'Iteration # ', IterationCount );
    write( OutFile, 'Desired Output:  ' );
    for i := 1 to OutSize do
      write( OutFile, Out_Pats[ Patt, i ]:6:3, '  ' );

    writeln( OutFile );
    write( OutFile, 'Actual Output:   ' );
    for i := 1 to OutSize do
      write( OutFile, Out_Out[ i ]:6:3, '  ' );
    writeln( OutFile );
  end; { Display_Output }

(*********************************************************
   do_forward_pass(patt)
     control function for the forward pass through the network
**********************************************************)
procedure Do_Forward_Pass( var Patt : byte; Matching : boolean );

  begin { Do_Forward_Pass }
    Do_Mid_Forward( Patt );     (* process forward pass, middle layer *)
    Do_Out_Forward;             (* process forward pass, output layer *)
    Display_Output( Patt, Matching );     (* display results of forward pass *)
  end; { Do_Forward_Pass }

(********************************************************
	do_out_error(patt)
		Compute the error for the output layer neurodes.
		This is simply Desired - Actual.
*********************************************************)
procedure Do_Out_Error( var Patt : byte );

  var
    Error, TotError : real{**};
    Neurode : integer;

  begin { Do_Out_Error }
    TotError := 0.0;
    for Neurode := 1 to OutSize do
      begin
        Out_Error[ Neurode ] := Out_Pats[ Patt, Neurode ] - Out_Out[ Neurode ];
        {Out_Error[ Neurode ] := Out_Pats[ Patt, Neurode ] -
                                round( Out_Out[ Neurode ] );}
	(* while we're here, also compute magnitude of total error for
	   output layer.  We use this to decide if we are done yet. *)
	Error := Out_Error[ Neurode ];
	if Error < 0 then
	  TotError := TotError - Error
	else
	  TotError := TotError + Error;
      end; { for }
    Tot_Out_Error[ Patt ] := TotError;
	                                (* used to check if network done *)
  end; { Do_Out_Error }

(*********************************************************
	do_mid_error()
		Compute the error for the middle layer neurodes
		This is based on the output errors computed above.
		Note that the derivative of the sigmoid f(x) is
			f'(x) = f(x)(1 - f(x))
		Recall that f(x) is merely the output of the middle
		layer neurode on the forward pass.
		
**********************************************************)
procedure Do_Mid_Error;

  var
    Sum : real{**};
    Neurode, i : integer;

  begin { Do_Mid_Error }
    for Neurode := 1 to MidSize do
      begin
	Sum := 0.0;
	for i := 1 to OutSize do
	  Sum := Sum + ( Out_Wts[ i, Neurode ] * Out_Error[i] );
	(* apply the derivative of the sigmoid here *)
        Mid_Error[ Neurode ] := Mid_Out[ Neurode ] *
                                ( 1 - Mid_Out[ Neurode ] ) * Sum;
      end; { for }
  end; { Do_Mid_Error }

(*********************************************************
	adjust_out_wts()
		Adjust the weights of the output layer.  The error for
		the output layer has been previously propagated back to
		the middle layer.
		Use the Delta Rule with momentum term to adjust the weights.
**********************************************************)
procedure Adjust_Out_Wts;

  var
    Length, Learn, Delta : real{**};
    Weight, Neurode : integer;

  begin { Adjust_Out_Wts }
    learn := BETA;
    length := 0.0;       (* since all output neurodes see the same input
                            pattern, only have to compute this once! *)
    for Weight := 1 to MidSize do
      Length := Length + ( Mid_Out[ Weight ] * Mid_Out[ Weight ] );
		(* since we really want the square of the length, we don't
		   bother taking the square root -- we'd only have to square
		   it again *)
    if Length <= 0.1 then
      Length := 0.1;

    for Neurode := 1 to OutSize do
      begin
        for Weight := 1 to MidSize do
	  begin
	    Delta := Learn * Out_Error[ Neurode ] * Mid_Out[ Weight ] / Length;
	    Out_Wts[ Neurode, Weight ] := Out_Wts[ Neurode, Weight ] + Delta;
	    Out_Wt_Change[ Neurode, Weight ] := Delta;
          end; { for }
      end; { for }
  end; { Adjust_Out_Wts }

(********************************************************
	adjust_mid_wts(patt)
		Adjust the middle layer weights using the previously computed
		errors.
		We use the Generalized Delta Rule with momentum term
*********************************************************)
procedure Adjust_Mid_Wts( var Patt : byte );

  var
    Length, Learn, Delta : real{**};
    Weight, neurode : integer;

  begin { Adjust_Mid_Wts }
    Learn := BETA;
    Length := 0.0;       (* since all output neurodes see the same input
                           pattern, only have to compute this once! *)
    for Weight := 1 to InSize do
      Length := Length + ( In_Pats[ Patt, Weight ] * In_Pats[ Patt, Weight ] );
		(* since we really want the square of the length, we don't
		   bother taking the square root -- we'd only have to square
		   it again *)
    if Length <= 0.1 then
      Length := 0.1;

    for Neurode := 1 to MidSize do
      begin
        for Weight := 1 to InSize do
	  begin
	    Delta := Learn * Mid_Error[ Neurode ] * In_Pats[ Patt, Weight ] / Length;
	    Mid_Wts[ Neurode, Weight ] := Mid_Wts[ Neurode, Weight ] + Delta;
	    Mid_Wt_change[ Neurode, Weight ] := Delta;
	  end; { for }
      end; { for }
  end; { Adjust_Mid_Wts }

(*********************************************************
	do_back_pass(patt)
		Process the backward propagation of error through network.
**********************************************************)
procedure Do_Back_Pass( var Patt : byte );

  begin { Do_Back_Pass }
    Do_Out_Error( Patt );
    Do_Mid_Error;
    Adjust_Out_Wts;
    Adjust_Mid_Wts( Patt );
  end; { Do_Back_Pass }

(********************************************************
	check_out_error()
		Check to see if the error in the output layer is below
		MARGIN*OUT_SIZE for all output patterns.  If so, then
		assume the network has learned acceptably well.  This 
		is simply an arbitrary measure of how well the network
		has learned -- many other better standards are possible.
*********************************************************)
function Check_Out_Error( var Error : boolean ) : boolean;

  var
    Standard, FinalErr : real{**};
    i : integer;

  begin { Check_Out_Error }
    Check_Out_Error := True;
    Error := False;
    FinalErr := 0.0;
    Standard := 0.8;
    writeln;
    writeln( 'Iteration # ', IterationCount );
    for i := 1 to NumPats do
      begin
        writeln( 'Total error for pattern ', i:2, ':  ', Tot_Out_Error[ i ]:8:3 );
        writeln( OutFile, 'Total error for pattern ', i:2, ':  ', Tot_Out_Error[ i ]:8:3 );
        FinalErr := FinalErr + Tot_Out_Error[i];
	if Tot_Out_Error[ i ] >= Standard then
          Check_Out_Error := False;
	if Tot_Out_Error[ i ] >= 16.0 then
          begin
            Error := True;
            Check_Out_Error := True;
          end;
      end; { for }
    writeln( 'Total error this pass thru input file:  ', FinalErr:6:3 );
    writeln( OutFile, 'Total error this pass thru input file:  ', FinalErr:6:3 );
    {Check_Out_Error := Result;}
  end; { Check_Out_Error }

(******************************************************
	randomize_wts()
		Intialize the weights in the middle and output layers to
		random values between -0.25..+0.25
*******************************************************)
procedure Randomize_Wts;

  var
    value : real{**};
    neurode, i : integer;

  begin { Randomize_Wts }

    writeln;
    write( 'Please enter a random number seed (1..32767):  ' );
    read( i );
    randseed := i;

    for neurode := 1 to MidSize do
      begin
        for i := 1 to InSize do
	  begin
	    value := ( random( 32768 ) / 32767.0 ) - 0.5;
	    mid_wts[ neurode, i ] := value / 2;
	    mid_wt_change[ neurode, i ] := 0.0;
	  end;
      end;

    for neurode := 1 to OutSize do
      begin
	for i := 1 to MidSize do
	  begin
            value := ( random( 32768 ) / 32767.0 ) - 0.5;
	    Out_Wts[ Neurode, i ] := Value / 2;
	    Out_Wt_Change[ Neurode, i ] := 0.0;
	  end; { for }
      end; { for }
  end; { Randomize_Wts }

(******************************************************
	read_data_file()
		Read in the input data file and store the patterns in
		in_pats and out_pats.
		
		The format for the data file is as follows:
		line#   data expected
		1		In-X-size,in-y-size,out-size
		2		number of patterns in file
		3		1st X row of 1st pattern
		4..		following rows of 1st pattern
		in-x+2	y-out pattern
		+1		1st X row of 2nd pattern
		etc.
		
		Each row of data is separated by commas or spaces.
		The data is expected to be ascii text corresponding to
		either a +1 or a 0.  
		
		Sample input for a 1-pattern file (The comments to the
		right may NOT be in the file unless more sophisticated
		parsing of the input is done.):
		
		5,7,8                   input is 5x7 grid, output is 8 bits
		1                       one pattern in file
		0,1,1,1,0               beginning of pattern for "O"
		1,0,0,0,1
		1,0,0,0,1
		1,0,0,0,1
		1,0,0,0,1
		1,0,0,0,0
		0,1,1,1,0
		0,1,0,0,1,1,1,1            ASCII code for "O" -- 0100 1111
		
		Clearly, this simple scheme can be expanded or enhanced
		any way you like.
		
		Returns -1 if any file error occurred, otherwise 0.
*******************************************************)
function Read_Error_On_Data_File : boolean;

  var
    InFile : text;
    InPath : string;
    XInSize,YInSize,YOutSize : integer;
    Patt, Element, i, Row, Col : integer;
    ch : char;
    Val1,Val2,Val3,Val4,Val5,Val6,Val7,Val8 : integer;

  begin { Read_Error_On_Data_File }
    InPath := paramstr( 2 );
    if InPath = '' then
      begin
        writeln;
        writeln( 'No input file specified.' );
        Read_Error_On_Data_File := True;
        exit;
      end;

    if CurrentMode = Match then
      OutWeight := 'NUMBERS' + OutWeight
    else
      OutWeight := InPath + OutWeight;

    writeln;
    writeln( 'Opening and retrieving data from file.' );

    {$I-}
    Assign( InFile, InPath );
    reset( InFile );
    {$I+}

    if IOresult <> 0 then
      begin
        writeln;
        writeln( 'Error in opening file!' );
        Read_Error_On_Data_File := True;
        exit;
      end;
    readln( InFile, XInSize, YInSize, YOutSize );
    readln( InFile, NumPats );
    if NumPats > MaxPats then
      NumPats := MaxPats;

    for Patt := 1 to NumPats do
      begin
	Element := 1;
	for Row := 1 to YInSize do
	  begin
            readln( InFile, Val1, Val2, Val3, Val4, Val5 );

	    Element := ( (Row-1) * XInSize ) + 1;
	    In_Pats[ Patt, Element ] := Val1;
            inc( Element );
	    In_Pats[ Patt, Element ] := Val2;
            inc( Element );
	    In_Pats[ Patt, Element ] := Val3;
            inc( Element );
	    In_Pats[ Patt, Element ] := Val4;
            inc( Element );
	    In_Pats[ Patt, Element ] := Val5;
            inc( Element );
	  end; { for }
	for i := 1 to InSize do
          begin
            if In_Pats[ Patt, i ] >= 0.9 then
              In_Pats[ Patt, i ] := 0.9;
	    if In_Pats[ Patt, i ] <= 0.1 then
	      In_Pats[ Patt, i ] := 0.1;
	  end;
	Element := 1;
        readln( InFile, Val1, Val2, Val3, Val4, Val5, Val6, Val7, Val8 );
	Out_Pats[ Patt, Element] := Val1;
        inc( Element );
	Out_Pats[ Patt, Element] := Val2;
        inc( Element );
	Out_Pats[ Patt, Element] := Val3;
        inc( Element );
	Out_Pats[ Patt, Element] := Val4;
        inc( Element );
        Out_Pats[ Patt, Element] := Val5;
        inc( Element );
        Out_Pats[ Patt, Element] := Val6;
        inc( Element );
        Out_Pats[ Patt, Element] := Val7;
        inc( Element );
	Out_Pats[ Patt, Element] := Val8;
        inc( Element );
      end; { for }

    writeln;
    writeln( 'Closing the input file now.' );
    close( InFile );
    Read_Error_On_Data_File := False;
  end; { Read_Data_File }

(*******************************************************
	initialize_net()
		Do all the initialization stuff before beginning
********************************************************)
function Initialize_Net_Error : boolean;

  begin { Initialize_Net_Error }
    Initialize_Net_Error := True;
    if not Read_Error_On_Data_File then
      begin
        Randomize_Wts;
        Initialize_Net_Error := False;
      end;
    IterationCount := 1;
  end; { Initialize_Net_Error }


(*******************************************************
	display_mid_wts()
		Display the weights on the middle layer neurodes
********************************************************)
procedure Display_Mid_Wts;

  var
    Neurode, Weight, Row, Col : integer;

  procedure Write_Weights;

    var
      Neurode, Weight, Row, Col : integer;

    begin { Write_Weights }
      {$I-}
      assign( WeightOut, OutWeight );
      rewrite( WeightOut );
      {$I+}
      if IOResult = 0 then
        for Neurode := 1 to MidSize do
          for row := 1 to InYSize do
            for col := 1 to InXSize do
	      begin
	        weight := InXSize * (row-1) + col;
                writeln( WeightOut, Mid_Wts[ Neurode, Weight ]:8:3 );
	      end; { for }
      close( WeightOut );
    end; { Write_Weights }

  begin { Display_Mid_Wts }
    writeln( OutFile );
    writeln( OutFile, 'Weights of Middle Layer neurodes:' );

    for Neurode := 1 to MidSize do
      begin
        writeln( OutFile );
        writeln( OutFile, '  Mid Neurode # ',neurode );
	for row := 1 to InYSize do
	  begin
            writeln( OutFile );
	    for col := 1 to InXSize do
	      begin
	        weight := InXSize * (row-1) + col;
                writeln( OutFile, Mid_Wts[ Neurode, Weight ]:8:3 );
	      end; { for }
	  end; { for }
      end; { for }

    if CurrentMode = Learn then
      Write_Weights;
    
  end; { Display_Mid_Wts }

(*******************************************************
	display_out_wts()
		Display the weights on the output layer neurodes
********************************************************)
procedure Display_Out_Wts;

  var
    Neurode, Weight : integer;

  procedure Write_Weights;

    var
      Neurode, Weight : integer;

    begin { Write_Weights }
      {$I-}
      append( WeightOut );
      {$I+}
      {if IOResult = 0 then}
      for Neurode := 1 to OutSize do
        for Weight := 1 to MidSize do
          writeln( WeightOut, Out_Wts[ Neurode, Weight ]:8:3 );
      close( WeightOut );
    end; { Write_Weights }

  begin { Display_Out_Wts }
    writeln( OutFile );
    writeln( OutFile, 'Weights of Output Layer neurodes:' );

    for Neurode := 1 to OutSize do
      begin
        writeln( OutFile );
        writeln( OutFile, '  Out Neurode # ', neurode );
        for Weight := 1 to MidSize do
          writeln( OutFile, Out_Wts[ Neurode, Weight ]:8:3 );
      end; { for }

    if CurrentMode = Learn then
      Write_Weights;

  end; { Display_Out_Wts }

function LearnAPattern : boolean;

  var
    Patt : byte;
    LearnError : boolean;

  begin { LearnAPattern }
    LearnError := False;
    CurrentMode := Learn;
    LearnAPattern := False;
    if Initialize_Net_Error then
      begin
        writeln;
        writeln;
        writeln( 'Program terminating with file error' );
        LearnAPattern := True;
        exit;
      end;
    
    Learned := False;
    while not Learned do
      begin
        for Patt := 1 to numpats do
          begin
            Do_Forward_Pass( Patt, False );
            Do_Back_Pass( Patt );
            inc( IterationCount );
            write( ' .' );
          end; { for }
        Learned := Check_Out_Error( LearnError );

        (* The following allows an easy bailout *)
	(* It terminates the program if the user has pressed any key *)
	(* at any time since the last pass through the file   *)
        if keypressed then
          Learned := True;
      end; { while }

      if LearnError then
        begin
          writeln;
          writeln;
          writeln( 'Error # ', learned, ' in learning after ', IterationCount, ' iterations' );
          writeln( OutFile );
          writeln( OutFile );
          writeln( OutFile, 'Error # ', learned, ' in learning after ', IterationCount, ' iterations' );
        end
      else
        begin
          writeln;
          writeln;
          writeln( 'TRAINING COMPLETE after ', IterationCount,' iterations' );
          writeln( OutFile );
          writeln( OutFile );
          writeln( OutFile, 'TRAINING COMPLETE after ', IterationCount,' iterations' );
        end;
    for Patt := 1 to Numpats do
      Do_Forward_Pass( Patt, False );
  end; { LearnAPattern }

function MatchAPattern : boolean;

  var
    Patt, i : byte;
    ch : char;

  function Load_Weights : boolean;

    var
      neurode, i : integer;

    begin { Load_Weights }
      Load_Weights := False;
      {$I-}
      Assign( WeightOut, OutWeight );
      Reset( WeightOut );
      {$I+}
      if IOResult <> 0 then
        begin
          Load_Weights := True;
          exit;
        end;
      for neurode := 1 to MidSize do
        for i := 1 to InSize do
	  begin
            readln( WeightOut, mid_wts[ neurode, i ] );
	    mid_wt_change[ neurode, i ] := 0.0;
	  end;
      for neurode := 1 to OutSize do
        for i := 1 to MidSize do
          begin
            readln( WeightOut, Out_Wts[ Neurode, i ] );
            Out_Wt_Change[ Neurode, i ] := 0.0;
	  end; { for }
      close( WeightOut );
    end; { Load_Weights }

  begin { MatchAPattern }
    CurrentMode := Match;
    MatchAPattern := True;
    if not Read_Error_On_Data_File then
      if not Load_Weights then
        MatchAPattern := False;
    writeln;
    writeln( '               Output' );
    {writeln( 'Pattern   Desired   Actual' );}
    writeln( '          Desired   Actual' );
    for Patt := 1 to NumPats do
      begin
        for i := 1 to InSize do
          begin
            if In_Pats[ Patt, i ] > 0.8 then
              write( '* ' )
            else
              write( '. ' );
            if ( i mod InXSize ) = 0 then
              writeln;
          end;
        Do_Forward_Pass( Patt, True );
        repeat
        until keypressed;
        ch := Readkey;
      end;
    writeln;
  end; { MatchAPattern }

(*****************************************************
******************************************************
		                  MAIN PROGRAM
******************************************************
******************************************************)
var
  ch : string;
  Error : boolean;

begin
  Error := True;
  OutWeight := '.OUT';
  assign( OutFile, OutPath );
  rewrite( OutFile );          (* open an output text file *)
  ch := paramstr( 1 );
  ch[ 1 ] := upcase( ch[ 1 ] );
  case ch[ 1 ] of
    'L' : Error := LearnAPattern;
    'M' : Error := MatchAPattern;
  else
    begin
      writeln( 'Syntax: BACKPROP mode inputpattern' );
      writeln( 'Valid mode parameters are: ' );
      writeln( '  L - Learn a pattern' );
      writeln( '  M - Match a pattern' );
      writeln;
      writeln( 'Inputpattern is the file of input patterns' );
      writeln;
      writeln( 'Please try again' );
      exit;
    end;
  end; { case }

  if not Error then
    if ch[ 1 ] <> 'M' then
      begin
        writeln( 'Posting final weights to file...' );
        Display_Mid_Wts;
        Display_Out_Wts;
      end;
  close( OutFile );
end.
